@@ -327,6 +327,25 @@ void bladeRF_SoapySDR::setGainMode(const int direction, const size_t channel, co
         SoapySDR::logf(SOAPY_SDR_ERROR, "bladerf_set_gain_mode(%s) returned %s", automatic?"automatic":"manual", _err2str(ret).c_str());
         throw std::runtime_error("setGainMode() " + _err2str(ret));
     }
+    bladerf_gain_mode return_mode;
+    bladerf_get_gain_mode(_dev, _toch(direction, channel), &return_mode);
+    std::string gain_mode_string;
+
+    if (return_mode == BLADERF_GAIN_DEFAULT) {
+        gain_mode_string = "default";
+    } else if (return_mode == BLADERF_GAIN_MGC) {
+        gain_mode_string = "manual";
+    } else if (return_mode == BLADERF_GAIN_FASTATTACK_AGC) {
+        gain_mode_string = "fastattack";
+    } else if (return_mode == BLADERF_GAIN_SLOWATTACK_AGC) {
+        gain_mode_string = "slowattack";
+    } else if (return_mode == BLADERF_GAIN_HYBRID_AGC) {
+        gain_mode_string = "hybrid";
+    } else {
+        gain_mode_string = "<unknown>";
+    }
+
+    SoapySDR::logf(SOAPY_SDR_INFO, "setGainMode(%s, %d, %d), actual = %s", direction==SOAPY_SDR_RX?"Rx":"Tx", int(channel), automatic, gain_mode_string.c_str());
 }
 
 bool bladeRF_SoapySDR::getGainMode(const int direction, const size_t channel) const
@@ -430,16 +449,82 @@ SoapySDR::Range bladeRF_SoapySDR::getGainRange(const int direction, const size_t
  * Frequency API
  ******************************************************************/
 
-void bladeRF_SoapySDR::setFrequency(const int direction, const size_t channel, const std::string &name, const double frequency, const SoapySDR::Kwargs &)
+void bladeRF_SoapySDR::setFrequency(const int direction, const size_t channel, const std::string &name, const double frequency, const SoapySDR::Kwargs &args)
 {
     if (name == "BB") return; //for compatibility
     if (name != "RF") throw std::runtime_error("setFrequency("+name+") unknown name");
 
+    //NOTE on quick tunes:
+    // - this is available on BladeRF2, not BladeRF1
+    // - there can be up to NUM_BBP_FASTLOCK_PROFILES (256) quick tunes in both directions.
+    //     bladerf2_get_quick_tune returns BLADERF_ERR_UNEXPECTED if you try to get more.
+    // - to clear previous quick tunes, the RFIC should have its state reset which is currently done
+    //     when the FPGA is loaded or reloaded. To have the ability to live reset the index to 0,
+    //     the ADI AXI core might have to be modified.
+
+    //if "saveQuickTune" == "1", set the frequency and store the quick tune parameter.
+    //NOTE that it's possible to overwrite previous quick tune parameter.
+    auto saveQuickTuneIter = args.find("saveQuickTune");
+    if (saveQuickTuneIter != args.end() && saveQuickTuneIter->second == "1")
+    {
+        if (!_isBladeRF2)
+        {
+            SoapySDR::logf(SOAPY_SDR_ERROR, "saveQuickTune is only available for BladeRF2.");
+            throw std::runtime_error("saveQuickTune is only available for BladeRF2.");
+        }
+
+        setRfFrequency(direction, channel, frequency);
+
+        auto quickTune = getQuickTune(direction, channel);
+        if (quickTune == nullptr)
+        {
+            SoapySDR::logf(SOAPY_SDR_ERROR, "Cannot set frequency for retune.");
+            throw std::runtime_error("Cannot set frequency for retune.");
+        }
+
+        auto quickTuneIter = _quickTunesByDirChanAndFreq.find({ direction, channel, frequency });
+        if (quickTuneIter != _quickTunesByDirChanAndFreq.end()) delete quickTuneIter->second;
+
+        _quickTunesByDirChanAndFreq[{direction, channel, frequency}] = quickTune;
+        return;
+    }
+
+    //Else, if "reuseQuickTune" == "1", we'll see if a corresponding quick tune parameter exists.
+    //If it does, we'll use it. Else, we throw an exception.
+    //If "timestamp" is specified, the quick tune will be scheduled at that time, else it'll be done at once.
+    auto reuseQuickTuneIter = args.find("reuseQuickTune");
+    if (reuseQuickTuneIter != args.end() && reuseQuickTuneIter->second == "1") {
+        if (!_isBladeRF2)
+        {
+            SoapySDR::logf(SOAPY_SDR_ERROR, "reuseQuickTune is only available for BladeRF2.");
+            throw std::runtime_error("reuseQuickTune is only available for BladeRF2.");
+        }
+
+        auto quickTuneIter = _quickTunesByDirChanAndFreq.find({ direction, channel, frequency });
+        if (quickTuneIter == _quickTunesByDirChanAndFreq.end())
+        {
+            SoapySDR::logf(SOAPY_SDR_ERROR, "Unkown quick tune for frequency %f and channel %d", frequency, channel);
+            throw std::runtime_error("Unkown quick tune");
+        }
+
+        auto value = args.find("timestamp");
+        long long timestamp = value == args.end() ? 0 : std::stoll(value->second);
+
+        retune(direction, channel, timestamp, quickTuneIter->second);
+        return;
+    }
+
+    //Else, we simply set the RF frequency.
+    setRfFrequency(direction, channel, frequency);
+}
+
+void bladeRF_SoapySDR::setRfFrequency(const int direction, const size_t channel, const double frequency)
+{
     int ret = bladerf_set_frequency(_dev, _toch(direction, channel), bladerf_frequency(std::round(frequency)));
     if (ret != 0)
     {
         SoapySDR::logf(SOAPY_SDR_ERROR, "bladerf_set_frequency(%f) returned %s", frequency, _err2str(ret).c_str());
-        throw std::runtime_error("setFrequency("+name+") " + _err2str(ret));
+        throw std::runtime_error("setFrequency(RF) " + _err2str(ret));
     }
 }
 
@@ -478,6 +563,35 @@ SoapySDR::RangeList bladeRF_SoapySDR::getFrequencyRange(const int direction, con
     return {toRange(range)};
 }
 
+bladerf_quick_tune* bladeRF_SoapySDR::getQuickTune(const int direction, const size_t channel) const
+{
+    bladerf_quick_tune* quick_tune = new bladerf_quick_tune();
+    bladerf_channel ch = _toch(direction, channel);
+    int ret = bladerf_get_quick_tune(_dev, ch, quick_tune);
+
+    if (ret != 0)
+    {
+        SoapySDR::logf(SOAPY_SDR_ERROR, "bladerf_get_quick_tune() returned %s", _err2str(ret).c_str());
+        delete quick_tune;
+        return nullptr;
+    }
+
+    return quick_tune;
+}
+
+void bladeRF_SoapySDR::retune(const int direction, const size_t channel, long long timestamp, bladerf_quick_tune* quickTune)
+{
+    bladerf_channel ch = _toch(direction, channel);
+
+    int ret = bladerf_schedule_retune(_dev, ch, timestamp, 0 /* frequency not needed for retune */, quickTune);
+
+    if (ret != 0)
+    {
+        SoapySDR::logf(SOAPY_SDR_ERROR, "bladerf_schedule_retune() returned %s", _err2str(ret).c_str());
+        throw std::runtime_error("retune() " + _err2str(ret));
+    }
+}
+
 /*******************************************************************
  * Sample Rate API
  ******************************************************************/
@@ -629,6 +743,95 @@ std::vector<double> bladeRF_SoapySDR::listBandwidths(const int direction, const
     return options;
 }
 
+/*******************************************************************
+ * Clocking API
+ ******************************************************************/
+
+void bladeRF_SoapySDR::setMasterClockRate(const double rate)
+{
+    if (! _isBladeRF2) return;
+
+    int ret = bladerf_set_pll_refclk(_dev, uint64_t(rate));
+
+    if (ret != 0)
+    {
+        SoapySDR::logf(SOAPY_SDR_ERROR, "bladerf_set_pll_refclk() returned %s", _err2str(ret).c_str());
+        throw std::runtime_error("setMasterClockRate() " + _err2str(ret));
+    }
+}
+
+double bladeRF_SoapySDR::getMasterClockRate(void) const
+{
+    if (! _isBladeRF2) return 0;
+
+    uint64_t rate(0);
+    int ret = bladerf_get_pll_refclk(_dev, &rate);
+
+    if (ret != 0)
+    {
+        SoapySDR::logf(SOAPY_SDR_ERROR, "bladerf_get_pll_refclk() returned %s", _err2str(ret).c_str());
+        throw std::runtime_error("getMasterClockRate() " + _err2str(ret));
+    }
+
+    return double(rate);
+}
+
+SoapySDR::RangeList bladeRF_SoapySDR::getMasterClockRates(void) const
+{
+    if (! _isBladeRF2) return SoapySDR::RangeList();
+
+    const bladerf_range* range(nullptr);
+    int ret = bladerf_get_pll_refclk_range(_dev, &range);
+
+    if (ret != 0)
+    {
+        SoapySDR::logf(SOAPY_SDR_ERROR, "bladerf_get_pll_refclk_range() returned %s", _err2str(ret).c_str());
+        throw std::runtime_error("getMasterClockRates() " + _err2str(ret));
+    }
+
+    return {toRange(range)};
+}
+
+std::vector<std::string> bladeRF_SoapySDR::listClockSources(void) const
+{
+    std::vector<std::string> clocks;
+    clocks.push_back("internal");
+    if (_isBladeRF2) clocks.push_back("ref_in");
+    return clocks;
+}
+
+void bladeRF_SoapySDR::setClockSource(const std::string &source)
+{
+    if (! _isBladeRF2) return;
+
+    bool enable = (source == "ref_in");
+    int ret = bladerf_set_pll_enable(_dev, enable);
+
+    if (ret != 0)
+    {
+        SoapySDR::logf(SOAPY_SDR_ERROR, "bladerf_set_pll_enable() returned %s", _err2str(ret).c_str());
+        throw std::runtime_error("setClockSource() " + _err2str(ret));
+    }
+}
+
+std::string bladeRF_SoapySDR::getClockSource(void) const
+{
+    if (! _isBladeRF2) return "internal";
+
+    bool enabled(false);
+    int ret = bladerf_get_pll_enable(_dev, &enabled);
+
+    if (ret != 0)
+    {
+        SoapySDR::logf(SOAPY_SDR_ERROR, "bladerf_get_pll_enable() returned %s", _err2str(ret).c_str());
+        throw std::runtime_error("getClockSource() " + _err2str(ret));
+    }
+
+    if (enabled) return "ref_in";
+    else return "internal";
+}
+
+
 /*******************************************************************
  * Time API
  ******************************************************************/
@@ -968,6 +1171,34 @@ SoapySDR::ArgInfoList bladeRF_SoapySDR::getSettingInfo(void) const
 
     setArgs.push_back(loadArg);
 
+    // BiasTee TX
+    SoapySDR::ArgInfo biasTeeTx;
+    biasTeeTx.key = "biastee_tx";
+    biasTeeTx.value = "false";
+    biasTeeTx.name = "Enable bias tee for tx";
+    biasTeeTx.description = "Enables the bias tee on the tx channel 0";
+    biasTeeTx.type = SoapySDR::ArgInfo::BOOL;
+    biasTeeTx.options.push_back("true");
+    biasTeeTx.optionNames.push_back("True");
+    biasTeeTx.options.push_back("false");
+    biasTeeTx.optionNames.push_back("False");
+
+    setArgs.push_back(biasTeeTx);
+
+    // BiasTee RX
+    SoapySDR::ArgInfo biasTeeRx;
+    biasTeeRx.key = "biastee_rx";
+    biasTeeRx.value = "false";
+    biasTeeRx.name = "Enable bias tee for rx";
+    biasTeeRx.description = "Enables the bias tee on the rx channel 0";
+    biasTeeRx.type = SoapySDR::ArgInfo::BOOL;
+    biasTeeRx.options.push_back("true");
+    biasTeeRx.optionNames.push_back("True");
+    biasTeeRx.options.push_back("false");
+    biasTeeRx.optionNames.push_back("False");
+
+    setArgs.push_back(biasTeeRx);
+
     return setArgs;
 }
 
@@ -999,6 +1230,10 @@ std::string bladeRF_SoapySDR::readSetting(const std::string &key) const
         return "false";
     } else if (key == "load_fpga") {
         return "";
+    } else if (key == "biastee_tx") {
+        return "false";
+    } else if (key == "biastee_rx") {
+        return "false";
     }
 
     SoapySDR_logf(SOAPY_SDR_WARNING, "Unknown setting '%s'", key.c_str());
@@ -1289,6 +1524,34 @@ void bladeRF_SoapySDR::writeSetting(const std::string &key, const std::string &v
             SoapySDR::logf(SOAPY_SDR_ERROR, "bladeRF: The provided FPGA image path is empty");
         }*/
     }
+    else if (key == "biastee_tx")
+    {
+        if (value == "true" || value == "false") {
+            // --> Valid setting has arrived
+            int ret = bladerf_set_bias_tee(_dev, BLADERF_CHANNEL_TX(0), value == "true");
+            if (ret != 0)
+            {
+                SoapySDR::logf(SOAPY_SDR_ERROR, "bladerf_set_bias_tee(BLADERF_CHANNEL_TX(0), %s) returned %s",
+                               value.c_str(),
+                               _err2str(ret).c_str());
+                throw std::runtime_error("writeSetting() " + _err2str(ret));
+            }
+        }
+    }
+    else if (key == "biastee_rx")
+    {
+        if (value == "true" || value == "false") {
+            // --> Valid setting has arrived
+            int ret = bladerf_set_bias_tee(_dev, BLADERF_CHANNEL_RX(0), value == "true");
+            if (ret != 0)
+            {
+                SoapySDR::logf(SOAPY_SDR_ERROR, "bladerf_set_bias_tee(BLADERF_CHANNEL_RX(0), %s) returned %s",
+                               value.c_str(),
+                               _err2str(ret).c_str());
+                throw std::runtime_error("writeSetting() " + _err2str(ret));
+            }
+        }
+    }
     else
     {
         throw std::runtime_error("writeSetting(" + key + ") unknown setting");
